#include "libmpeg2.h"
#include "mpeg2protos.h"

#include <stdlib.h>
#include <string.h>

int mpeg2_new(mpeg2_t **file)
{
	*file = malloc(sizeof(mpeg2_t));
	(*file)->packet_size = 0;
	(*file)->has_audio = 0;
	(*file)->has_video = 0;
	(*file)->atrack = 0;
	(*file)->vtrack = 0;
	(*file)->is_transport_stream = 0;
	(*file)->is_program_stream = 0;
	(*file)->is_audio_stream = 0;
	(*file)->is_video_stream = 0;
}

int mpeg2_delete(mpeg2_t *file)
{
	if(file->vtrack)
	{
		mpeg2_delete_vtrack(file, file->vtrack);
	}

	if(file->atrack)
	{
		mpeg2_delete_atrack(file, file->atrack);
	}

	free(file);
}

int mpeg2_check_sig(char *path)
{
	mpeg2_fs_t fs;
	unsigned int bits;
	char *ext;
	int result = 0;

	if(mpeg2io_open_file(&fs, path))
	{
// File not found
		return 0;
	}

	bits = mpeg2io_read_int32(&fs);
// Test header
	if((((bits >> 24) & 0xff) == MPEG2_SYNC_BYTE) ||
		(bits == MPEG2_PACK_START_CODE) ||
		((bits & 0xfff00000) == 0xfff00000) ||
		(bits == MPEG2_SEQUENCE_START_CODE))
	{
		result = 1;

		ext = strrchr(path, '.');
		if(ext)
		{
// Test file extension.
			if(strncasecmp(ext, ".mp2", 4) && 
				strncasecmp(ext, ".mp3", 4) &&
				strncasecmp(ext, ".m2v", 4) &&
				strncasecmp(ext, ".m2s", 4) &&
				strncasecmp(ext, ".mpg", 4) &&
				strncasecmp(ext, ".vob", 4))
				result = 0;
		}
	}

	mpeg2io_close_file(&fs);
	return result;
}

mpeg2_t* mpeg2_open(char *path)
{
	mpeg2_t *file = 0;
	unsigned int bits;

// Initialize the file structure
	mpeg2_new(&file);
	if(mpeg2io_open_file(&(file->fs), path))
	{
		mpeg2_delete(file);
		return 0;
	}

// Figure out what streams are available
	bits = mpeg2io_read_int32(&(file->fs));

	if(((bits >> 24) & 0xff) == MPEG2_SYNC_BYTE)
	{
// Transport stream with audio and video
		file->packet_size = MPEG2_TS_PACKET_SIZE;
		file->has_audio = 1;
		file->has_video = 1;
		file->is_transport_stream = 1;
	}
	else
	if(bits == MPEG2_PACK_START_CODE)
	{
// Program stream with audio and video.
		file->packet_size = MPEG2_DVD_PACKET_SIZE;
		file->has_audio = 1;
		file->has_video = 1;
		file->is_program_stream = 1;
	}
	else
	if((bits & 0xfff00000) == 0xfff00000)
	{
// Audio only
		file->packet_size = 2048;
		file->has_audio = 1;
		file->is_audio_stream = 1;
	}
	else
	if(bits == MPEG2_SEQUENCE_START_CODE)
	{
// Video only
		file->packet_size = 2048;
		file->has_video = 1;
		file->is_video_stream = 1;
	}
	else
	{
		mpeg2_delete(file);
		fprintf(stderr, "mpeg2_open: not an MPEG 2 stream\n");
		return 0;
	}
	file->total_bytes = mpeg2io_total_bytes(&(file->fs));

// Create structures for all data types
	if(file->has_video)
	{
		file->vtrack = mpeg2_new_vtrack(file);
		if(!file->vtrack) file->has_video = 0;
	}
// Video must be created before audio because audio uses the video timecode
// to get its length.
	if(file->has_audio)
	{
		file->atrack = mpeg2_new_atrack(file);
		if(!file->atrack) file->has_audio = 0;
	}

	mpeg2io_close_file(&(file->fs));
	return file;
}

int mpeg2_close(mpeg2_t *file)
{
// File is closed in the same procedure it is opened in.
	mpeg2_delete(file);
	return 0;
}

int mpeg2_has_audio(mpeg2_t *file)
{
	return file->has_audio;
}

int mpeg2_audio_channels(mpeg2_t *file)
{
	if(file->has_audio)
		return file->atrack->channels;
	return -1;
}

int mpeg2_sample_rate(mpeg2_t *file)
{
	if(file->has_audio)
		return file->atrack->sample_rate;
	return -1;
}

long mpeg2_get_sample(mpeg2_t *file)
{
	if(file->has_audio)
		return file->atrack->current_position;
	return -1;
}

int mpeg2_set_sample(mpeg2_t *file, long sample)
{
	if(file->has_audio)
	{
		file->atrack->current_position = sample;
		return 0;
	}
	return -1;
}

long mpeg2_audio_samples(mpeg2_t *file)
{
	if(file->has_audio)
		return file->atrack->total_samples;
	return -1;
}

int mpeg2_has_video(mpeg2_t *file)
{
	return file->has_video;
}

int mpeg2_video_width(mpeg2_t *file)
{
	if(file->has_video)
		return file->vtrack->width;
	return -1;
}

int mpeg2_video_height(mpeg2_t *file)
{
	if(file->has_video)
		return file->vtrack->height;
	return -1;
}

float mpeg2_frame_rate(mpeg2_t *file)
{
	if(file->has_video)
		return file->vtrack->frame_rate;
	return -1;
}

long mpeg2_video_frames(mpeg2_t *file)
{
	if(file->has_video)
		return file->vtrack->total_frames;
	return -1;
}

long mpeg2_get_frame(mpeg2_t *file)
{
	if(file->has_video)
		return file->vtrack->current_position;
	return -1;
}

int mpeg2_set_frame(mpeg2_t *file, long frame)
{
	if(file->has_video)
	{
		file->vtrack->current_position = frame;
		return 0;
	}
	return -1;
}

int mpeg2_read_frame(mpeg2_t *file, unsigned char *output)
{
	if(file->has_video && file->vtrack->current_position < file->vtrack->total_frames)
	{
		int result;
		result = mpeg2video_read_frame(file->vtrack->video, file->vtrack->current_position, output);
		file->vtrack->current_position++;
		return result;
	}
	return -1;
}

int mpeg2_read_audio(mpeg2_t *file, float *output_f, short *output_i, int channel, long samples)
{
	if(file->has_audio && file->atrack->current_position < file->atrack->total_samples)
	{
		int result;
		result = mpeg2audio_decode_audio(file->atrack->audio, output_f, output_i, channel, file->atrack->current_position, samples);
		file->atrack->current_position += samples;
		return result;
	}
	return -1;
}

int mpeg2_read_audio_chunk(mpeg2_t *file, unsigned char *output, long *size, long max_size)
{
	int result = 0;
	if(file->has_audio)
	{
		result = mpeg2audio_read_raw(file->atrack->audio, output, size, max_size);
	}
	return result;
}

int mpeg2_read_video_chunk(mpeg2_t *file, unsigned char *output, long *size, long max_size)
{
	int result = 0;
	if(file->has_video)
	{
		result = mpeg2video_read_raw(file->vtrack->video, output, size, max_size);
	}
	return result;
}
