// Concatenate elementary streams

#include "libmpeg2.h"

#include <stdio.h>
#include <stdlib.h>

#define MPEG2_SEQUENCE_START_CODE        0x000001b3
#define BUFFER_SIZE            1000000

int main(int argc, char *argv[])
{
	char inpath[1024];
	mpeg2_t *in;
	int current_file;
	unsigned int bits;
	unsigned char *buffer;
	long output_size;
	int result = 0;
	long total_frames = 0;

	if(argc < 2)
	{
		fprintf(stderr, "mpeg2cat concatenates elementary MPEG streams and writes them to stdout.\n"
			"Usage: mpeg2cat <infile> [infile...] > <outfile>\n"
			"Example: mpeg2cat xena1.m2v xena2.m2v > xena.m2v\n"
			"         mpeg2cat xena1.mp2 xena2.mp2 > xena.mp2\n");
		exit(1);
	}


	buffer = malloc(BUFFER_SIZE);

	for(current_file = 1; current_file < argc; current_file++)
	{
		strcpy(inpath, argv[current_file]);
		if(!(in = mpeg2_open(inpath)))
		{
			fprintf(stderr, "Skipping %s\n", inpath);
			continue;
		}

		if(mpeg2_has_audio(in))
		{
// Add audio stream to end
			while(!mpeg2_read_audio_chunk(in, buffer, &output_size, BUFFER_SIZE))
			{
				result = !fwrite(buffer, output_size, 1, stdout);
				if(result)
				{
					perror("fwrite audio chunk");
					break;
				}
			}
		}
		else
		if(mpeg2_has_video(in))
		{
// Add video stream to end
			int hour, minute, second, frame;
			long gop_frame;
			unsigned long code;
			float carry;
			int i;

			while(!mpeg2_read_video_chunk(in, buffer, &output_size, BUFFER_SIZE) &&
				output_size >= 4)
			{
// Increment the frame count.
				code = (unsigned long)buffer[output_size - 4] << 24; 
				code |= (unsigned long)buffer[output_size - 3] << 16; 
				code |= (unsigned long)buffer[output_size - 2] << 8; 
				code |= (unsigned long)buffer[output_size - 1]; 

// Got a frame
				if(code == MPEG2_PICTURE_START_CODE)
				{
					total_frames++;
				}
				else
				if(code == MPEG2_SEQUENCE_END_CODE)
				{
// Clip the sequence end code.
					output_size -= 4;
				}

// Fix the GOP header if there is ones
				i = 0;
				code = (unsigned long)buffer[i++] << 24;
				code |= (unsigned long)buffer[i++] << 16;
				code |= (unsigned long)buffer[i++] << 8;
				code |= buffer[i++];
				while(i < output_size && code != MPEG2_GOP_START_CODE && code != MPEG2_SEQUENCE_END_CODE)
				{
					code <<= 8;
					code |= buffer[i++];
				}
				
				if(code == MPEG2_GOP_START_CODE)
				{
// Get the time code
					code = (unsigned long)buffer[i] << 24;
					code |= (unsigned long)buffer[i + 1] << 16;
					code |= (unsigned long)buffer[i + 2] << 8;
					code |= (unsigned long)buffer[i + 3];
					
					hour = code >> 26 & 0x1f;
					minute = code >> 20 & 0x3f;
					second = code >> 13 & 0x3f;
					frame = code >> 7 & 0x3f;

					gop_frame = (long)(hour * 3600 * mpeg2_frame_rate(in) +
							minute * 60 * mpeg2_frame_rate(in) +
							second * mpeg2_frame_rate(in) + 
							frame);
fprintf(stderr, "old: %02d:%02d:%02d:%02d ", hour, minute, second, frame);
// Write a new time code
//					if(gop_frame != total_frames - 1)
					{
						hour = (long)((float)(total_frames - 1) / mpeg2_frame_rate(in) / 3600);
						carry = hour * 3600 * mpeg2_frame_rate(in);
						minute = (long)((float)(total_frames - 1 - carry) / mpeg2_frame_rate(in) / 60);
						carry += minute * 60 * mpeg2_frame_rate(in);

						second = (long)((float)(total_frames - 1 - carry) / mpeg2_frame_rate(in));
						carry += second * mpeg2_frame_rate(in);
						frame = (total_frames - 1 - carry);
						buffer[i] = ((code >> 24) & 0x80) | (hour << 2) | (minute >> 4);
						buffer[i + 1] = ((code >> 16) & 0x08) | ((minute & 0xf) << 4) | (second >> 3);
						buffer[i + 2] = ((second & 0x7) << 5) | (frame >> 1);
						buffer[i + 3] = (code & 0x7f) | ((frame & 0x1) << 7);
// 					code = (unsigned long)buffer[i] << 24;
// 					code |= (unsigned long)buffer[i + 1] << 16;
// 					code |= (unsigned long)buffer[i + 2] << 8;
// 					code |= (unsigned long)buffer[i + 3];
fprintf(stderr, "new: %02d:%02d:%02d:%02d\r", hour, minute, second, frame);
					}
				}

// Write the frame
				result = !fwrite(buffer, output_size, 1, stdout);
				if(result)
				{
					perror("fwrite video chunk");
					break;
				}
			}
		}
		else
		{
			fprintf(stderr, "Unsupported stream type.\n");
			mpeg2_close(in);
			continue;
		}
		
		mpeg2_close(in);
	}

// Terminate output
	if(mpeg2_has_video(in))
	{
fprintf(stderr, "\n");
// Write new end of sequence
		buffer[0] = MPEG2_SEQUENCE_END_CODE >> 24;
		buffer[1] = (MPEG2_SEQUENCE_END_CODE >> 16) & 0xff;
		buffer[2] = (MPEG2_SEQUENCE_END_CODE >> 8) & 0xff;
		buffer[3] = MPEG2_SEQUENCE_END_CODE & 0xff;
		result = !fwrite(buffer, 4, 1, stdout);
	}

	exit(0);
}
