#include "audiodevice.h"
#include "errorbox.h"
#include "file.h"
#include "record.h"
#include "recordaudio.h"
#include "recordgui.h"
#include "recordengine.h"
#include "renderengine.h"


RecordAudio::RecordAudio(Record *record, 
				RecordEngine *engine, 
				AudioDevice *device,
				int write_file)
 : Thread()
{
	this->record = record;
	this->engine = engine; 
	this->device = device;
	this->write_file = write_file;
	is_recording = 0;
	is_paused = 0;
	//completed.unlock();
	//monitor_paused.unlock();
	fragment_position = 0;
	total_samples = 0;
	quit_when_completed = 0;
}

RecordAudio::~RecordAudio()
{
}

int RecordAudio::set_parameters(File *file,
							RecordGUI *gui,
							BC_Meter **meter, 
							long fragment_size, 
							long buffer_size, 
							int record_channels, 
							long *dc_offset, 
							int realtime)
{
	this->file = file;
	this->gui = gui;
	this->meter = meter;
	this->fragment_size = fragment_size;
	if(realtime) set_realtime();

	for(this->buffer_size = fragment_size; this->buffer_size < buffer_size; this->buffer_size += fragment_size)
		;

	this->record_channels = record_channels;
	this->dc_offset = dc_offset;
// dithering is done at the device for recording and depends on the bits of the output file
	device->set_record_dither(engine->get_dither());
	if(!file) write_file = 0;
}

int RecordAudio::start_recording()
{     // limitations in the sound driver cause spurts of data 
	if(!is_recording)
	{
		is_paused = 0;
		if(!write_file)
		{
// Monitor mode
			engine->open_audio_input();
			pause_lock.lock();
			synchronous = 1;
		}
		else
		{
// Record mode
// Let the thread quit on its own.
			synchronous = 0;
			completed.lock();
		}

		is_recording = 1;        // flag thread to start when it's ready
		Thread::start();
	}
}

int RecordAudio::stop_recording()
{
	if(is_recording) 
	{
		is_recording = 0;         // thread unpauses monitor
		if(!write_file)
		{
			Thread::join();
			pause_lock.unlock();
			engine->close_audio_input();
		}
		else
		{
// 			completed.lock();      // wait for recording completion
// 			completed.unlock();
		}
	}
}

int RecordAudio::wait_for_completion()
{
	completed.lock();      // wait for recording completion
	completed.unlock();
}

int RecordAudio::pause_recording()
{
	if(is_recording && !write_file)
	{
		is_paused = 1;
		pause_lock.lock();
		engine->close_audio_input();
	}
}

int RecordAudio::resume_recording()
{
	if(is_recording && !write_file)
	{
		is_paused = 0;
		total_samples = 0;
		engine->open_audio_input();
		monitor_paused.unlock();
	}
}

void RecordAudio::run()
{
	int channel, buffer;
	int *over, over_count;
	float *max;
	int result = 0;
	Timer delayer;
	int total_clipped_samples = 0;
	int clipped_sample = 0;
	int debug = 0;

	total_samples = 0;

	max = new float[record_channels];
	over = new int[record_channels];

	for(channel = 0; channel < record_channels; channel++)
	{
		device->set_dc_offset(dc_offset[channel], channel);
	}

	fragment_position = 0;

// thread out I/O
	if(write_file)
	{
// Get a buffer from the file to record into.
		file->start_audio_thread(buffer_size);
		input = file->get_audio_buffer();
	}
	else
	{
// Make up a fake buffer.
		input = new float*[record_channels];

		for(int i = 0; i < record_channels; i++)
		{
			input[i] = new float[buffer_size];
		}
	}

	gui->total_clipped_samples = 0;
	gui->update_clipped_samples(0);

if(debug) printf("RecordAudio::run 2 %d %d\n", is_recording, result);
	while(is_recording && !result)
	{
if(debug) printf("RecordAudio::run 3\n");
		if(!write_file) monitor_paused.lock();
if(debug) printf("RecordAudio::run 4\n");
debug = 0;

// Handle data from the audio device.
		if(record->monitor_audio || write_file)
		{
			if(write_file)
			{
// Read into file's buffer for recording.
// device needs to write buffer starting at fragment position
				device->read_buffer(input, fragment_size, record_channels, over, max, fragment_position);
			}
			else
			{
// Read into monitor buffer for monitoring.
				device->read_buffer(input, fragment_size, record_channels, over, max, 0);
			}

// Update timer for synchronization
			timer_lock.lock();
			total_samples += fragment_size;
			timer.update();
			timer_lock.unlock();

// Update meters if monitoring
			if(record->monitor_audio && is_recording && !result)
			{
				engine->lock_window();
				for(channel = 0; channel < record_channels; channel++)
				{
					meter[channel]->update(max[channel], over[channel]);
				}

				engine->unlock_window();
			}

			if(record->monitor_audio || write_file)
			{
				clipped_sample = 0;
				for(channel = 0; channel < record_channels; channel++)
				{
					if(over[channel]) clipped_sample = 1;
				}
			}

// Write file if writing
			if(write_file)
			{
				fragment_position += fragment_size;

				if(fragment_position >= buffer_size)
				{
					result = write_buffer();
				}

				engine->update_position(get_position() + fragment_size);
				if(clipped_sample) gui->update_clipped_samples(++total_clipped_samples);

// handle different recording modes
				if(engine->get_record_mode() == 2 && engine->current_position > engine->get_loop_duration() && !result)
				{
// loop
					set_position(0);
				}
				else
				if(result || (engine->get_record_mode() == 1 && engine->current_position > engine->get_loop_duration()))
				{
// timed record and the timer has run out or write error
					is_recording = 0;
					engine->is_saving = 0;
				}
			}
		}
		else
		{
// Don't handle audio data.
			delayer.delay(250);
			timer_lock.lock();
			total_samples += record->get_samplerate() / 4;
			timer.update();
			timer_lock.unlock();
		}

// Handle pausing 
		if(!write_file)
		{
			if(!is_paused) 
			{
				monitor_paused.unlock();
			}
			else 
			{
				pause_lock.unlock();
			}
		}
	}

if(debug) printf("RecordAudio::run 5\n");
	if(result)
	{
		is_recording = 0;
		ErrorBox error_box;
		error_box.create_objects("No space left on disk.");
		error_box.run_window();
	}

	if(write_file)
	{
// write last buffer
		result = file->write_audio_buffer(fragment_position);
// stop file I/O
		file->stop_audio_thread();
	}
	else
	{
// Delete the fake buffer.
		for(int i = 0; i < record_channels; i++)
		{
			meter[i]->reset();
			delete input[i];
		}
		delete input;
		input = 0;
	}

// reset meter
	engine->lock_window();
	for(channel = 0; channel < record_channels; channel++)
	{
		meter[channel]->reset();
	}

	engine->unlock_window();
	delete max;
	delete over;

// duplex stops here
	if(write_file)
	{
		engine->stop_duplex();
	}

	engine->close_audio_input();

	if(write_file)
	{
		if(!record->do_video) engine->resume_monitor(1, 0);     // resume monitor.  Failed to stop video.
		completed.unlock();
		if(record->do_video) engine->resume_monitor(1, 0);     // resume monitor.  Stopped video but crashed.
// The use of locking in the audio recorder is contingent on the use of a video recorder.
// The video recorder's lock takes precidence over the audio recorder's lock.
	}

// 	if(quit_when_completed)
// 	{
// 		quit_when_completed = 0;
// 		gui->set_done(0);
// 	}
}

int RecordAudio::write_buffer()
{
// block until buffer is ready for writing
	int result = file->write_audio_buffer(fragment_position);
	fragment_position = 0;
	input = file->get_audio_buffer();
	return result;
}

int RecordAudio::set_position(long position)
{
	int result = file->write_audio_buffer(fragment_position);
	fragment_position = 0;

// must stop the thread to relocate
	file->stop_audio_thread();
	file->set_audio_position(position);
	engine->update_position(0);
	
	file->start_audio_thread(buffer_size);
	
	input = file->get_audio_buffer();
	return result;	
}

long RecordAudio::get_position()
{
	return engine->current_position;
}

long RecordAudio::absolute_position()
{
	long result;
	if(is_recording)
	{
		timer_lock.lock();
		result = total_samples;
		result += (long)((float)timer.get_difference() / 1000 * record->get_samplerate());
		timer_lock.unlock();
		return result;
	}
	else
	return -1;
}


int RecordAudio::set_dc_offset(long new_offset, int number)
{
	device->set_dc_offset(new_offset, number);
}

long int RecordAudio::get_dc_offset(long *dc_offset, RecordGUIDCOffsetText **dc_offset_text)
{
	return device->get_dc_offset(dc_offset, dc_offset_text);
}
