#include "errorbox.h"
#include "file.h"
#include "record.h"
#include "recordengine.h"
#include "recordgui.h"
#include "recordvideo.h"
#include "recvideowindow.h"
#include "vframe.h"
#include "videodevice.h"

#include <unistd.h>

RecordVideo::RecordVideo(Record *record, 
	RecordEngine *engine, 
	VideoDevice *vdevice, 
	int write_file)
 : Thread()
{
	this->record = record;
	this->vdevice = vdevice;
	this->write_file = write_file;
	this->engine = engine;
	is_recording = 0;
	is_paused = 0;
	this->write_file = write_file;
	quit_when_completed = 0;
}

RecordVideo::~RecordVideo()
{
}

int RecordVideo::set_parameters(File *file,
							RecordGUI *gui,
							int buffer_size,
							int realtime)
{
	this->file = file;
	this->gui = gui;
	this->buffer_size = buffer_size;
	if(realtime) set_realtime();
	if(!file) write_file = 0;
}

int RecordVideo::start_recording()
{
	if(!is_recording)
	{
		is_paused = 0;
		total_frames = 0;
		if(!write_file)
		{
			engine->open_video_input();
			pause_lock.lock();
			synchronous = 1;
		}
		else
		{
			synchronous = 0;
			completed.lock();
		}

		is_recording = 1;
		Thread::start();
	}
}

int RecordVideo::stop_recording()
{
	if(is_recording)
	{
		is_recording = 0;
		if(engine->vdevice->get_failed())
		{
			Thread::end();
			Thread::join();
			pause_lock.unlock();
			engine->close_video_input();

			if(write_file)
			{
				if(!record->do_audio) engine->stop_duplex();
				file->write_video_buffer(buffer_position, 0, record->use_floatingpoint());
				file->stop_video_thread();
				engine->resume_monitor(0, 1);
				completed.unlock();
			}
		}
		else
		{
			if(!write_file)
			{
				Thread::join();
				pause_lock.unlock();
				engine->close_video_input();
			}
			else
			{
// 				completed.lock();
// 				completed.unlock();
// Recording thread closes automatically
			}
		}
	}
}

int RecordVideo::wait_for_completion()
{
	if(!engine->vdevice->get_failed())
	{
		completed.lock();
		completed.unlock();
	}
}

int RecordVideo::pause_recording()
{
	if(is_recording)
	{
		is_paused = 1;
		pause_lock.lock();
		engine->close_video_input();
	}
}

int RecordVideo::resume_recording()
{
	if(is_recording)
	{
		total_frames = 0;
		is_paused = 0;
		engine->open_video_input();
		monitor_paused.unlock();
	}
}

void RecordVideo::run()
{
	VFrame *capture_frame;
	Timer delayer;
	int result = 0;
	VFrame ***frame_ptr;
	long current_sample;   // Sample in time of the start of the capture
	long next_sample;      // Sample of next frame
	long total_dropped_frames = 0;  // Total number of frames behind
	long dropped_frames = 0;  // Number of frames currently behind
	long last_dropped_frames = 0;  // Number of dropped frames during the last calculation
	long delay;
	long record_start = 0;     // Frame start of this recording in the file

// Thread out the I/O
	if(write_file)
	{
		record_start = file->get_video_position(record->get_frame_rate());
		file->start_video_thread(buffer_size, 0);
		frame_ptr = file->get_video_buffer();
		buffer_position = 0;
	}
	else
	{
		capture_frame = new VFrame(record->frame_w, record->frame_h, 0);
		frame_ptr = new VFrame**[1];
		frame_ptr[0] = new VFrame*[1];
		frame_ptr[0][0] = capture_frame;
	}

	gui->total_dropped_frames = 0;  // Number of frames for user to know about.
	gui->update_dropped_frames(0);

	while(is_recording && !result)
	{
// Merely pauses when switching from monitor to record
		if(!write_file) 
			monitor_paused.lock();
		else
		if(buffer_position >= buffer_size)
		{
			frame_ptr = file->get_video_buffer();
			buffer_position = 0;
		}

// ============================= Synchronize with audio or timer
		if(is_recording && (record->monitor_video || write_file))
//		if(is_recording && write_file)
		{
			next_sample = (long)((float)total_frames / record->get_framerate() * record->get_samplerate());

 			if(write_file) 
				current_sample = engine->absolute_record_position();
 			else
 				current_sample = engine->absolute_monitor_position();

			if(current_sample < next_sample && current_sample > 0)
			{
// Too early.
				delay = (long)((float)(next_sample - current_sample) / record->get_samplerate() * 1000);
// Sanity check and delay.
				if(delay < 10000 && delay > 0) delayer.delay(delay);
			}
			else
			if(current_sample > 0 && write_file)
			{
// Captured too slow.
				dropped_frames = (long)((float)(current_sample - next_sample) / record->get_samplerate() * record->get_framerate());
// Add the frames currently behind to the total dropped frames
				if(dropped_frames != last_dropped_frames)
				{
//					total_dropped_frames += dropped_frames - last_dropped_frames;
					gui->update_dropped_frames(dropped_frames);
					last_dropped_frames = dropped_frames;
				}
//				last_dropped_frames = dropped_frames;
			}
		}

// ================================ Capture a frame
		if(record->monitor_video || write_file)
		{

			if(write_file)
			{
				capture_frame = frame_ptr[0][buffer_position];
				record->video_lock.lock();
				vdevice->read_buffer(capture_frame);
				record->video_lock.unlock();
				buffer_position++;
			}
			else
			{
				vdevice->read_buffer(capture_frame);
			}

			if(write_file && !record->do_audio) 
				engine->update_position(tosamples(record_start + total_frames, 
					record->get_samplerate(), 
					record->get_framerate()));
		}

// Flash the frame
		if(record->monitor_video)
			gui->monitor_video_window->update(capture_frame);

// Compress the frames
		if(write_file && buffer_position >= buffer_size)
		{
			result = file->write_video_buffer(buffer_position, 0, record->use_floatingpoint());
		}

// Delay if no monitoring and no compressing
		if(!record->monitor_video && !write_file)
			delayer.delay(250);

		if(write_file)
		{
// handle different recording modes
			if(engine->get_record_mode() == 2 && engine->current_position > engine->get_loop_duration() && !result)
			{
// loop
// must stop the thread to relocate
				result = file->write_video_buffer(buffer_position, 0, record->use_floatingpoint());
				file->stop_video_thread();
				file->set_video_position(0, record->get_frame_rate());
				file->start_video_thread(buffer_size, 0);
				buffer_position = 0;
				record_start = -total_frames - 1;
				if(!record->do_audio) engine->update_position(0);
			}
			else
			if(result || (engine->get_record_mode() == 1 && engine->current_position > engine->get_loop_duration()))
			{
// timed record and the timer has run out
				is_recording = 0;
				engine->is_saving = 0;
			}
		}

// =============================== Test paused condition.
		if(!is_paused) 
		{
			monitor_paused.unlock();
		}
		else 
		{
			pause_lock.unlock();
		}
// Advance to next frame
		total_frames++;
	}

	if(result)
	{
		is_recording = 0;
		if(!record->do_audio)
		{
			ErrorBox error_box;
			error_box.create_objects("No space left on disk.");
			error_box.run_window();
		}
	}

	engine->close_video_input();
	if(write_file)
	{
		if(!record->do_audio) engine->stop_duplex();
// write last buffer
		result = file->write_video_buffer(buffer_position, 0, record->use_floatingpoint());
// stop file I/O
		file->stop_video_thread();
		engine->resume_monitor(0, 1);
		completed.unlock();
	}
	else
	{
		delete frame_ptr[0];
		delete frame_ptr;
		delete capture_frame;
	}
	
// 	if(!record->do_audio && quit_when_completed)
// 	{
// 		quit_when_completed = 0;
// 		gui->set_done(0);
// 	}
}

int RecordVideo::unhang_thread()
{
	printf("RecordVideo::unhang_thread\n");
	Thread::end();
}

