/* 
 *
 *	mantissa.c Copyright (C) Aaron Holtzman - May 1999
 *
 *
 *  This file is part of libmpeg2
 *	
 *  libmpeg2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  libmpeg2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "mpeg2audio.h"


//Lookup tables of 0.16 two's complement quantization values
unsigned short mpeg2_q_1[3] = {( -2 << 15)/3, 0           ,(  2 << 15)/3 };

unsigned short mpeg2_q_2[5] = {( -4 << 15)/5,( -2 << 15)/5,   0         ,
	                (  2 << 15)/5,(  4 << 15)/5};

unsigned short mpeg2_q_3[7] = {( -6 << 15)/7,( -4 << 15)/7,( -2 << 15)/7,
	                   0         ,(  2 << 15)/7,(  4 << 15)/7,
									(  6 << 15)/7};

unsigned short mpeg2_q_4[11] = {(-10 << 15)/11,(-8 << 15)/11,(-6 << 15)/11,
	                ( -4 << 15)/11,(-2 << 15)/11,  0          ,
									(  2 << 15)/11,( 4 << 15)/11,( 6 << 15)/11,
									(  8 << 15)/11,(10 << 15)/11};

unsigned short mpeg2_q_5[15] = {(-14 << 15)/15,(-12 << 15)/15,(-10 << 15)/15,
                   ( -8 << 15)/15,( -6 << 15)/15,( -4 << 15)/15,
									 ( -2 << 15)/15,   0          ,(  2 << 15)/15,
									 (  4 << 15)/15,(  6 << 15)/15,(  8 << 15)/15,
									 ( 10 << 15)/15,( 12 << 15)/15,( 14 << 15)/15};

static unsigned short mpeg2_qnttztab[16] = { 0, 0, 0, 3, 0 , 4, 5, 6, 7, 8, 9, 10, 11, 12, 14, 16};

void mpeg2audio_ac3_mantissa_reset(mpeg2_ac3_mantissa_t *mantissa)
{
	mantissa->m_1[2] = mantissa->m_1[1] = mantissa->m_1[0] = 0;
	mantissa->m_2[2] = mantissa->m_2[1] = mantissa->m_2[0] = 0;
	mantissa->m_4[1] = mantissa->m_4[0] = 0;
	mantissa->m_1_pointer = mantissa->m_2_pointer = mantissa->m_4_pointer = 3;
}

/* 
 * Generate eight bits of pseudo-entropy using a 16 bit linear
 * feedback shift register (LFSR). The primitive polynomial used
 * is 1 + x^4 + x^14 + x^16.
 *
 * The distribution is uniform, over the range [-0.707,0.707]
 *
 */
unsigned int mpeg2audio_ac3_dither_gen(mpeg2audio_t *audio)
{
	int i;
	unsigned int state;

// explicitly bring the state into a local var as gcc > 3.0?
// doesn't know how to optimize out the stores
	state = audio->ac3_lfsr_state;

//Generate eight pseudo random bits
	for(i = 0; i < 8; i++)
	{
		state <<= 1;	

		if(state & 0x10000)
			state ^= 0xa011;
	}

	audio->ac3_lfsr_state = state;

	return (((((int)state << 8) >> 8) * (int)(0.707106f * 256.0f)) >> 16);
}


/* Fetch an unpacked, left justified, and properly biased/dithered mantissa value */
unsigned short mpeg2audio_ac3_mantissa_get(mpeg2audio_t *audio, 
	unsigned short bap, 
	unsigned short dithflag)
{
	unsigned short result;
	unsigned short group_code;
	mpeg2_ac3_mantissa_t *mantissa = &(audio->ac3_mantissa);

//If the bap is 0-5 then we have special cases to take care of
	switch(bap)
	{
		case 0:
			if(dithflag)
				result = mpeg2audio_ac3_dither_gen(audio);
			else
				result = 0;
			break;

		case 1:
			if(mantissa->m_1_pointer > 2)
			{
				group_code = mpeg2bits_getbits(audio->astream, 5);

				if(group_code > 26)
					//FIXME do proper block error handling
					printf("\n!! Invalid mantissa !!\n");

				mantissa->m_1[0] = group_code / 9; 
				mantissa->m_1[1] = (group_code % 9) / 3; 
				mantissa->m_1[2] = (group_code % 9) % 3; 
				mantissa->m_1_pointer = 0;
			}
			result = mantissa->m_1[mantissa->m_1_pointer++];
			result = mpeg2_q_1[result];
			break;
		case 2:

			if(mantissa->m_2_pointer > 2)
			{
				group_code = mpeg2bits_getbits(audio->astream, 7);

				if(group_code > 124)
					//FIXME do proper block error handling
					printf("\n!! Invalid mantissa !!\n");

				mantissa->m_2[0] = group_code / 25;
				mantissa->m_2[1] = (group_code % 25) / 5 ;
				mantissa->m_2[2] = (group_code % 25) % 5 ; 
				mantissa->m_2_pointer = 0;
			}
			result = mantissa->m_2[mantissa->m_2_pointer++];
			result = mpeg2_q_2[result];
			break;

		case 3:
			result = mpeg2bits_getbits(audio->astream, 3);

			if(result > 6)
				//FIXME do proper block error handling
				printf("\n!! Invalid mantissa !!\n");

			result = mpeg2_q_3[result];
			break;

		case 4:
			if(mantissa->m_4_pointer > 1)
			{
				group_code = mpeg2bits_getbits(audio->astream, 7);

				if(group_code > 120)
					//FIXME do proper block error handling
					printf("\n!! Invalid mantissa !!\n");

				mantissa->m_4[0] = group_code / 11;
				mantissa->m_4[1] = group_code % 11;
				mantissa->m_4_pointer = 0;
			}
			result = mantissa->m_4[mantissa->m_4_pointer++];
			result = mpeg2_q_4[result];
			break;

		case 5:
			result = mpeg2bits_getbits(audio->astream, 4);

			if(result > 14)
				//FIXME do proper block error handling
				printf("\n!! Invalid mantissa !!\n");

			result = mpeg2_q_5[result];
			break;

		default:
			result = mpeg2bits_getbits(audio->astream, mpeg2_qnttztab[bap]);
			result <<= 16 - mpeg2_qnttztab[bap];
	}

	return result;
}

int mpeg2audio_ac3_mantissa_unpack(mpeg2audio_t *audio, 
	mpeg2_ac3bsi_t *bsi, 
	mpeg2_ac3audblk_t *audblk)
{
	int i, j;
	int done_cpl = 0;

	mpeg2audio_ac3_mantissa_reset(&(audio->ac3_mantissa));

	for(i = 0; i < bsi->nfchans; i++)
	{
		for(j = 0; j < audblk->endmant[i]; j++)
			audblk->chmant[i][j] = mpeg2audio_ac3_mantissa_get(audio, audblk->fbw_bap[i][j], audblk->dithflag[i]);

		if(audblk->cplinu && audblk->chincpl[i] && !(done_cpl))
		{
/* ncplmant is equal to 12 * ncplsubnd */
/* Don't dither coupling channel until channel separation so that
 * interchannel noise is uncorrelated */
			for(j = audblk->cplstrtmant; j < audblk->cplendmant; j++)
				audblk->cplmant[j] = mpeg2audio_ac3_mantissa_get(audio, audblk->cpl_bap[j], 0);
			done_cpl = 1;
		}
	}

	if(bsi->lfeon)
	{
/* There are always 7 mantissas for lfe, no dither for lfe */
		for(j = 0; j < 7 ; j++)
			audblk->lfemant[j] = mpeg2audio_ac3_mantissa_get(audio, audblk->lfe_bap[j], 0);
	}
	return 0;
}
