#include "vframe.h"
#include <stdio.h>

VFrame::VFrame(int w, int h, int alpha)
{
	reset_flags();
	allocate_uncompressed_data(0, w, h, alpha);
}

VFrame::VFrame(VPixel *data, int w, int h, int alpha)
{
	reset_flags();
	allocate_uncompressed_data(data, w, h, alpha);
}

VFrame::VFrame(unsigned char *data, int w, int h)
{
	reset_flags();
	shared = 1;
	this->w = w;
	this->h = h;

	rows_noalpha = data;
}

VFrame::VFrame()
{
	reset_flags();
}

VFrame::~VFrame()
{
	clear_objects();
}

int VFrame::clear_objects()
{
	if(shared)
	{
		delete rows;
	}
	else
	{
		if(has_alpha) 
		{
			delete rows[0];
			delete rows;
		}
		else 
			delete rows_noalpha;
		
		if(compressed_data) delete compressed_data;
	}
}

int VFrame::reset_flags()
{
	compressed_data = 0;	  // With compressed data.
	compressed_allocated = 0;  // Allocated space
	compressed_size = 0;   // Size of current image
	shared = 0;
	has_alpha = 0;
	rows = 0;
	rows_noalpha = 0;
	w = 0;
	h = 0;
}

// Reallocate uncompressed buffer with or without alpha
int VFrame::reallocate(VPixel *data, int w, int h, int alpha)
{
	clear_objects();
	reset_flags();
	allocate_uncompressed_data(data, w, h, alpha);
}

int VFrame::allocate_uncompressed_data(VPixel *data, int w, int h, int alpha)
{
	this->w = w;
	this->h = h;
	this->has_alpha = alpha;

	if(data)
	{
		shared = 1;
		rows = new VPixel*[h];
		for(int i = 0; i < h; i++)
		{
			rows[i] = &data[i * w];
		}
	}
	else
	{
		shared = 0;
		if(alpha)
		{
			VPixel *total = new VPixel[w * h];

			rows = new VPixel*[h];
			for(int i = 0; i < h; i++)
			{
				rows[i] = &total[i * w];
			}
		}
		else
		{
			rows_noalpha = new unsigned char[w * h * 3];
		}
	}
}


int VFrame::allocate_compressed_data(long bytes)
{
	if(bytes < 1) return 1;
	if(compressed_data && compressed_allocated < bytes)
	{
		delete compressed_data;
		compressed_data = 0;
	}
	if(!compressed_data)
	{
		compressed_data = new char[bytes];
		compressed_allocated = bytes;
		compressed_size = 0;
	}
	return 0;
}

char* VFrame::get_compressed_data()
{
	return compressed_data;
}

long VFrame::get_compressed_allocated()
{
	return compressed_allocated;
}

long VFrame::get_compressed_size()
{
	return compressed_size;
}

long VFrame::set_compressed_size(long size)
{
	compressed_size = size;
	return 0;
}

int VFrame::is_alpha()
{
	return has_alpha;
}


int VFrame::equals(VFrame *frame)
{
	if(frame->rows[0] == rows[0]) 
	return 1;
	else
	return 0;
}

int VFrame::clear_frame()
{
	if(has_alpha)
	{
		for(register int i = 0, j; i < h; i++)
		{
			for(j = 0; j < w; j++)
			{
				clear_pixel(&rows[i][j]);
			}
		}
	}
	else
	{
		for(register int i = 0; i < h * w * 3; i++)
		{
				rows_noalpha[i] = 0;
		}
	}
}

int VFrame::apply_fade(int alpha)
{
	register int i, j, channel;
	if(alpha != VMAX)
	{
		for(i = 0; i < h; i++)
		{
			for(j = 0; j < w; j++)
			{
				channel = rows[i][j].alpha;
				channel *= alpha;
				channel >>= sizeof(VWORD) * 8;
				rows[i][j].alpha = (VWORD)channel;
			}
		}
	}
}

int VFrame::replace_from(VFrame *frame, int alpha, int use_alpha, int use_float)
{
	register long i, j;


	if(use_alpha && alpha != VMAX)
	{
		float a_float;
		int a_int, channel;

		if(use_float)
		{
			a_float = (float)alpha / VMAX;
		}
		else
		{
			a_int = alpha;
#if (VMAX == 65535)
			a_int >>= 8;
#endif
		}
		
		for(i = 0; i < h; i++)
		{
			if(use_float)
			{
				for(j = 0; j < w; j++)
				{
					rows[i][j].red = frame->rows[i][j].red;
					rows[i][j].green = frame->rows[i][j].green;
					rows[i][j].blue = frame->rows[i][j].blue;
					rows[i][j].alpha = (VWORD)(rows[i][j].alpha * a_float);
				}
			}
			else
			{
				for(j = 0; j < w; j++)
				{
					rows[i][j].red = frame->rows[i][j].red;
					rows[i][j].green = frame->rows[i][j].green;
					rows[i][j].blue = frame->rows[i][j].blue;
					rows[i][j].alpha = (VWORD)(((int)rows[i][j].alpha * a_int) >> 8);
				}
			}
		}
	}
	else
	{
		for(i = 0; i < h; i++)
		{
			for(j = 0; j < w; j++)
			{
				rows[i][j] = frame->rows[i][j];
			}
		}
	}
}

int VFrame::copy_from(VFrame *frame)
{
	long total = w * h;
	VPixel *block2 = frame->get_rows()[0];
	VPixel *block1 = get_rows()[0];

	for(long i = 0; i < total; i++)
	{
		block1[i] = block2[i];
	}
}


int VFrame::get_scale_tables(int *column_table, int *row_table, 
			int in_x1, int in_y1, int in_x2, int in_y2,
			int out_x1, int out_y1, int out_x2, int out_y2)
{
	int y_out, i;
	float w_in = in_x2 - in_x1;
	float h_in = in_y2 - in_y1;
	int w_out = out_x2 - out_x1;
	int h_out = out_y2 - out_y1;

	float hscale = w_in / w_out;
	float vscale = h_in / h_out;

	for(i = 0; i < w_out; i++)
	{
		column_table[i] = (int)(hscale * i);
	}

	for(i = 0; i < h_out; i++)
	{
		row_table[i] = (int)(vscale * i) + in_y1;
	}
}

int VFrame::clear_pixel(VPixel *pixel)
{
	pixel->red = pixel->green = pixel->blue = pixel->alpha = 0;
}

VPixel** VFrame::get_rows()
{
	if(rows)
		return rows;
	else
	{
		printf("VFrame::get_rows: no alpha channel\n");
		return 0;
	}
}

unsigned char* VFrame::get_rows_noalpha()
{
	if(rows_noalpha)
		return rows_noalpha;
	else
	{
		printf("VFrame::get_rows_noalpha: attempted to return an alpha channel\n");
		return 0;
	}
}


int VFrame::get_w()
{
	return w;
}

int VFrame::get_h()
{
	return h;
}

