#include "settings.h"
#include "asset.h"
#include "defaults.h"
#include "errorbox.h"
#include "file.h"
#include "filesystem.h"
#include "mainwindow.h"
#include "playbackengine.h"


MainWindow::MainWindow(ArrayList<char*> *playlist)
{
	this->playlist = playlist;
	engine = new PlaybackEngine(this);
	defaults = new Defaults("~/.xmovierc");
	defaults->load();
	reset_parameters();
	error_thread = new ErrorThread;
}

MainWindow::~MainWindow()
{
	delete error_thread;
	delete defaults;
	delete gui;
	if(asset) delete asset;
	if(file) delete file;
	if(frame) delete frame;
	reset_parameters();
}

int MainWindow::reset_parameters()
{
	file = 0;
	asset = 0;
	frame = 0;
	samplerate = 44100;
	frame_rate = 10;
	total_samples = 44100;
	current_sample = 0;
}

int MainWindow::create_objects()
{
	sprintf(default_path, "");
	defaults->get("PATH", default_path);
	every_frame = defaults->get("EVERYFRAME", 0);
	aspect_w = defaults->get("ASPECTW", (float)4);
	aspect_h = defaults->get("ASPECTH", (float)3);
	smp = defaults->get("SMP", 0);
	square_pixels = defaults->get("ASPECTSQUARE", 0);
	software_sync = defaults->get("SOFTWARESYNC", 0);
	prebuffer_size = defaults->get("PREBUFFER_SIZE", 2000000);
	gui = new MWindowGUI(this, 
					defaults->get("WIDTH", DEFAULTW), 
					defaults->get("HEIGHT", DEFAULTH));
	gui->create_objects();
}

int MainWindow::run_program()
{
	gui->run_window();
}

int MainWindow::load_file(char *path)
{
	close_file();
	asset = new Asset(path);
	file = new File;


	file->set_prebuffer(prebuffer_size);
	file->set_processors(smp ? 2 : 1);
	int result = file->try_to_open_file(asset);

	if(!result)
	{
// success
		if(asset->audio_data) samplerate = asset->rate;
		if(asset->video_data) frame_rate = asset->frame_rate;
		total_samples = get_length();
		current_sample = 0;

		if(asset->video_data)
		{
// resize frame
			int w = asset->width, h = asset->height;
			fix_size(w, h, gui->zoom_factor);
			gui->resize_canvas(w, h);
		}

// 		if(load_into_memory)
// 		{
// 			gui->canvas->notify_load();
// 			file->load_into_ram();
// 		}
// 

		FileSystem fs;
		char string1[1024], string2[1024];
		fs.extract_name(string1, asset->path);
		sprintf(string2, "XMovie: %s", string1);
		gui->set_title(string2);
		gui->resize_scrollbar();
		if(asset->video_data) flash_frame(current_sample);
		gui->update_scrollbar();
		if(asset->video_data) gui->canvas->flash();
		gui->menu->add_load(path);
		return 0;
	}
	else
	{
// failure
		switch(result)
		{
			case 1:
				error_thread->run_error("No such file or directory.");
				break;
			case 2:
				error_thread->run_error("Unsupported file format.");
				break;
			case 3:
				error_thread->run_error("The file contains no supported codecs.");
				break;
		}
		delete file;
		delete asset;
		asset = 0;
		file = 0;
		return 1;
	}
	return 0;
}

long MainWindow::get_length()
{
	long test_length1 = 0, test_length2 = 0;
	if(asset->video_data)
		test_length1 = (long)((float)file->get_video_length() / frame_rate * samplerate);

	if(asset->audio_data)
		test_length2 = file->get_audio_length();

	if(test_length1 > test_length2) return test_length1; else return test_length2;
}

int MainWindow::fix_size(int &w, int &h, float zoom_factor)
{
	int proper_w, proper_h;

	if(asset && asset->video_data)
	{
		if(asset->width * aspect_h > asset->height * aspect_w)
		{
			proper_w = (int)((float)w * zoom_factor);
			proper_h = (int)((float)w * zoom_factor / get_aspect_ratio());
		}
		else
		{
			proper_h = (int)((float)h * zoom_factor);
			proper_w = (int)((float)h * zoom_factor * get_aspect_ratio());
		}
		w = proper_w;
		h = proper_h;
	}
}

float MainWindow::get_aspect_ratio()
{
	if(square_pixels && asset)
	return (float)asset->width / asset->height;
	else
	return aspect_w / aspect_h;
}

int MainWindow::flash_frame(long frame_number)
{
	if(asset && asset->video_data)
	{
		if(!frame) this->frame = new unsigned char[asset->width * asset->height * 3];

		file->lock_read();
		file->set_video_position(frame_number);
		file->read_frame(this->frame);
		file->unlock_read();
		gui->flash_frame(this->frame);
	}
}

int MainWindow::arm_frame(long frame_number)
{
	if(asset && asset->video_data)
	{
		unsigned char **frame_buffer = gui->bitmap->get_row_pointers();
		int color_model = gui->bitmap->get_color_model();
		if(file->frame_buffer_copy_possible(color_model))
		{
			file->lock_read();
			file->set_video_position(frame_number);
			file->read_frame_buffer(frame_buffer, gui->bitmap->get_w(), gui->bitmap->get_h(), color_model);
			file->unlock_read();
		}
		else
		{
			if(!frame) this->frame = new unsigned char[asset->width * asset->height * 3];

			file->lock_read();
			file->set_video_position(frame_number);
			file->read_frame(this->frame);
			file->unlock_read();
		}
	}
}

int MainWindow::flash_armed_frame()
{
	if(asset && asset->video_data)
	{
		if(file->frame_buffer_copy_possible(gui->bitmap->get_color_model()))
		{
			gui->flash_frame(0);
		}
		else
		{
			if(this->frame)
				gui->flash_frame(this->frame);
		}
	}
}

int MainWindow::close_file()
{
	if(file)
	{
		file->close_file();
	}
	if(asset) delete asset;
	if(file) delete file;
	if(frame) delete frame;
	reset_parameters();
}

int MainWindow::exit_cleanly()
{
	close_file();
	gui->menu->save_loads(defaults);
	defaults->update("EVERYFRAME", every_frame);
	defaults->update("ASPECTW", aspect_w);
	defaults->update("ASPECTH", aspect_h);
	defaults->update("SMP", smp);
	defaults->update("ASPECTSQUARE", square_pixels);
	defaults->update("PATH", default_path);
	defaults->update("SOFTWARESYNC", software_sync);
	defaults->update("FRAMESCALE", gui->zoom_factor);
	defaults->update("PREBUFFER_SIZE", prebuffer_size);
	defaults->save();
}

int MainWindow::start_playback()
{
	if(asset)
		engine->start_playback();
}

int MainWindow::stop_playback()
{
	if(asset)
		engine->stop_playback();
	
	gui->playbutton->update(1);
}

int MainWindow::frame_forward()
{
	if(asset)
		if((long)toframes(current_sample, samplerate, frame_rate) <
			(long)toframes(total_samples, samplerate, frame_rate))
		{
			current_sample += (long)(samplerate / frame_rate + 0.5);
			flash_frame(toframes(current_sample, samplerate, frame_rate));
			gui->canvas->flash();
			gui->update_scrollbar();
		}
}

int MainWindow::frame_backward()
{
	if(asset)
	{
		long frame = (long)toframes(current_sample, samplerate, frame_rate) - 1;
		current_sample = tosamples(frame, samplerate, frame_rate);
		if(current_sample < 0) current_sample = 0;
		flash_frame(toframes(current_sample, samplerate, frame_rate));

		gui->canvas->flash();
		gui->update_scrollbar();
	}
}




MWindowGUI::MWindowGUI(MainWindow *mwindow, int w, int h)
 : BC_Window("XMovie", w, h, 25, 25, 1, 0)
{
	this->mwindow = mwindow;
	bitmap = 0;
}

MWindowGUI::~MWindowGUI()
{
	if(bitmap) delete bitmap;
}


int MWindowGUI::create_objects()
{
	int x = 0, y = 0;
	add_tool(menu = new MainMenu(mwindow, get_w()));
	menu->create_objects();
	y += menu->get_h();
	add_tool(canvas = new MainCanvas(mwindow, x, y, 
		mwindow->defaults->get("WIDTH", DEFAULTW), 
		mwindow->defaults->get("HEIGHT", DEFAULTH - BUTTONH - y)));

	y += canvas->get_h();

	add_tool(playbutton = new PlayButton(mwindow, x, y));
	x += playbutton->get_w() + 5;
	add_tool(scrollbar = new MainScrollbar(mwindow, x, y, get_w() - BUTTONW * 3 - 10));
	x += scrollbar->get_w() + 5;
	add_tool(backbutton = new FrameBackButton(mwindow, x, y));
	x += backbutton->get_w();
	add_tool(forwardbutton = new FrameForwardButton(mwindow, x, y));

	zoom_factor = mwindow->defaults->get("FRAMESCALE", (float)1);
}

int MWindowGUI::close_event()
{
	mwindow->exit_cleanly();
	set_done(0);
}

int MWindowGUI::resize_canvas(int w, int h)
{
	if(canvas->get_w() != w
		|| canvas->get_h() != h)
	{
		resize_window(w, h + BUTTONH + menu->get_h());
		canvas->set_size(canvas->get_x(), canvas->get_y(), 
			w, h);
	}
	resize_widgets();
}

int MWindowGUI::flash_frame(unsigned char* frame)
{
	if(bitmap && 
		(bitmap->get_w() != canvas->get_w() ||
		bitmap->get_h() != canvas->get_h()))
	{
		delete bitmap;
		bitmap = 0;
	}

	if(!bitmap)
	{
		bitmap = canvas->new_bitmap(canvas->get_w(), canvas->get_h());
	}

	if(frame)
		bitmap->read_frame(frame, mwindow->asset->width, mwindow->asset->height);

	canvas->draw_bitmap(bitmap, 1);
}

int MWindowGUI::resize_event(int w, int h)
{
// fix dimensions for canvas size
	if(mwindow->asset && mwindow->asset->video_data)
	{
		if(mwindow->asset->width * mwindow->aspect_h >= mwindow->asset->height * mwindow->asset->width)
		{
			zoom_factor = (float)w / mwindow->asset->width;
			if(abs(w - mwindow->asset->width) < 50) zoom_factor = 1;
		}
		else
		{
			zoom_factor = (float)w / mwindow->get_aspect_ratio() / mwindow->asset->height;
			if(abs((int)(w - mwindow->asset->height * mwindow->get_aspect_ratio())) < 50) zoom_factor = 1;
		}

//		zoom_factor = (float)w / mwindow->asset->width;
		w = mwindow->asset->width;
		h = mwindow->asset->height;
		mwindow->fix_size(w, h, zoom_factor);

		//if(zoom_factor > 0.5 && zoom_factor < 1.5) zoom_factor = 1;
//		if(abs(w - mwindow->asset->width) < 50) zoom_factor = 1;
// 
// 		w = (int)(mwindow->asset->width * zoom_factor);
// 		h = (int)(mwindow->asset->height * zoom_factor);
		zoom_factor = (float)w / mwindow->asset->width;
	}
	else
	{
		h -= menu->get_h() + BUTTONH;
	}

	resize_window(w, h + menu->get_h() + BUTTONH);
	canvas->set_size(canvas->get_x(), canvas->get_y(), w, h);
	resize_widgets();
	mwindow->flash_frame(toframes(mwindow->current_sample, mwindow->samplerate, mwindow->frame_rate));
	canvas->flash();
}

int MWindowGUI::resize_widgets()
{
	int y = get_h() - BUTTONH;
	int x = 0;
	playbutton->resize_tool(x, y);
	x += BUTTONW + 5;
	scrollbar->resize_tool(x, y, get_w() - BUTTONW * 3 - 10, BUTTONH);
	x += scrollbar->get_w() + 5;
	backbutton->resize_tool(x, y);
	x += BUTTONW;
	forwardbutton->resize_tool(x, y);
	resize_scrollbar();
	flash();
}

int MWindowGUI::resize_scrollbar()
{
	scrollbar->set_position(get_w() - BUTTONW * 3 - 10, mwindow->current_sample, 0, mwindow->total_samples);
}

int MWindowGUI::update_scrollbar()
{
	scrollbar->update(mwindow->current_sample);
}

MainMenu::MainMenu(MainWindow *mwindow, int w)
 : BC_MenuBar(0, 0, w)
{
	this->mwindow = mwindow;
}

MainMenu::~MainMenu()
{
}

int MainMenu::create_objects()
{
	add_menu(filemenu = new BC_Menu("File"));
	filemenu->add_menuitem(load_file = new Load(mwindow));
	//filemenu->add_menuitem(new Close(mwindow));
	filemenu->add_menuitem(new BC_MenuItem("-"));
	filemenu->add_menuitem(new Quit(mwindow));
	init_loads(mwindow->defaults);

	BC_Menu *optionsmenu;
	add_menu(optionsmenu = new BC_Menu("Options"));
	optionsmenu->add_menuitem(new EveryFrame(mwindow));
//	optionsmenu->add_menuitem(new LoadIntoMemory(mwindow));
	optionsmenu->add_menuitem(mwindow->settingsmenu = new SettingsMenu(mwindow));
	optionsmenu->add_menuitem(new OriginalSize(mwindow));
	optionsmenu->add_menuitem(new SoftwareSync(mwindow));
}


int MainMenu::init_loads(Defaults *defaults)
{
	total_loads = defaults->get("TOTAL_LOADS", 0);
	char string[1024], path[1024], filename[1024];
	FileSystem dir;
	if(total_loads > 0) filemenu->add_menuitem(new BC_MenuItem("-"));

	for(int i = 0; i < total_loads; i++)
	{
		sprintf(string, "LOADPREVIOUS%d", i);
		defaults->get(string, path);

		filemenu->add_menuitem(load_previous[i] = new LoadPrevious(load_file));
		dir.extract_name(filename, path);
		load_previous[i]->set_text(filename);
		load_previous[i]->set_path(path);
	}
}

int MainMenu::save_loads(Defaults *defaults)
{
	defaults->update("TOTAL_LOADS", total_loads);
	char string[1024];
	for(int i = 0; i < total_loads; i++)
	{
		sprintf(string, "LOADPREVIOUS%d", i);
		defaults->update(string, load_previous[i]->path);
	}
}

int MainMenu::add_load(char *path)
{
	if(total_loads == 0)
	{
		filemenu->add_menuitem(new BC_MenuItem("-"));
	}
	
// test for existing copy
	FileSystem fs;
	char text[1024], new_path[1024];      // get text and path
	fs.extract_name(text, path);
	strcpy(new_path, path);
	
	for(int i = 0; i < total_loads; i++)
	{
		if(!strcmp(load_previous[i]->text, text))     // already exists
		{                                // swap for top load
			for(int j = i; j > 0; j--)   // move preceeding loads down
			{
				load_previous[j]->set_text(load_previous[j - 1]->text);
				load_previous[j]->set_path(load_previous[j - 1]->path);
			}
			load_previous[0]->set_text(text);
			load_previous[0]->set_path(new_path);
			
			return 1;
		}
	}
	
// add another load
	if(total_loads < TOTAL_LOADS)
	{
		filemenu->add_menuitem(load_previous[total_loads] = new LoadPrevious(load_file));
		total_loads++;
	}
	
// cycle loads down
	for(int i = total_loads - 1; i > 0; i--)
	{
	// set menu item text
		load_previous[i]->set_text(load_previous[i - 1]->text);
	// set filename
		load_previous[i]->set_path(load_previous[i - 1]->path);
	}

// set up the new load
	load_previous[0]->set_text(text);
	load_previous[0]->set_path(new_path);
	return 0;
}

LoadPrevious::LoadPrevious(Load *loadfile)
 : BC_MenuItem("")
{ this->loadfile = loadfile; }

int LoadPrevious::handle_event()
{
// run as thread to prompt users about missing files
	loadfile->mwindow->gui->unlock_window();
	loadfile->mwindow->stop_playback();
	loadfile->mwindow->gui->lock_window();
	loadfile->mwindow->close_file();
	loadfile->mwindow->load_file(path);
}

int LoadPrevious::set_path(char *path)
{
	strcpy(this->path, path);
}

Quit::Quit(MainWindow *mwindow)
 : BC_MenuItem("Quit", "q", 'q')
{
	this->mwindow = mwindow;
}

Quit::~Quit()
{
}

int Quit::handle_event()
{
	mwindow->gui->unlock_window();
	mwindow->stop_playback();
	mwindow->gui->lock_window();
	mwindow->exit_cleanly();
	mwindow->gui->set_done(0);
}

Load::Load(MainWindow *mwindow)
 : BC_MenuItem("Open...", "o", 'o')
{
	this->mwindow = mwindow;
	thread = new LoadThread(mwindow);
}

Load::~Load()
{
	delete thread;
}

int Load::handle_event()
{
	mwindow->gui->unlock_window();
	mwindow->stop_playback();
	mwindow->gui->lock_window();
	thread->start();
	//thread->run(); // debugging
}
// 
// LoadIntoMemory::LoadIntoMemory(MainWindow *mwindow)
//  : BC_MenuItem("Load into RAM")
// {
// 	set_checked(mwindow->load_into_memory);
// 	this->mwindow = mwindow;
// }
// LoadIntoMemory::~LoadIntoMemory()
// {
// }
// int LoadIntoMemory::handle_event()
// {
// 	mwindow->load_into_memory ^= 1;
// 	set_checked(mwindow->load_into_memory);
// }

SoftwareSync::SoftwareSync(MainWindow *mwindow)
 : BC_MenuItem("Synchronize using software")
{
	set_checked(mwindow->software_sync);
	this->mwindow = mwindow;
}
SoftwareSync::~SoftwareSync()
{
}
int SoftwareSync::handle_event()
{
	mwindow->software_sync ^= 1;
	set_checked(mwindow->software_sync);
}

OriginalSize::OriginalSize(MainWindow *mwindow)
 : BC_MenuItem("Original Size")
{
	this->mwindow = mwindow;
}
OriginalSize::~OriginalSize()
{
}
int OriginalSize::handle_event()
{
	mwindow->gui->zoom_factor = 1;

	if(mwindow->asset && mwindow->asset->video_data)
	{
		int w = mwindow->asset->width, h = mwindow->asset->height;
		mwindow->fix_size(w, h, mwindow->gui->zoom_factor);
		mwindow->gui->resize_canvas(w, h);
	}
	else
		mwindow->gui->resize_canvas(mwindow->gui->canvas->get_w(), mwindow->gui->canvas->get_h());

	mwindow->flash_frame(toframes(mwindow->current_sample, mwindow->samplerate, mwindow->frame_rate));
	mwindow->gui->canvas->flash();
}

SettingsMenu::SettingsMenu(MainWindow *mwindow)
 : BC_MenuItem("Settings...")
{
	this->mwindow = mwindow;
	thread = new SettingsThread(mwindow);
}
SettingsMenu::~SettingsMenu()
{
	delete thread;
}
int SettingsMenu::handle_event()
{
	thread->start();
}

EveryFrame::EveryFrame(MainWindow *mwindow)
 : BC_MenuItem("Play every frame")
{
	this->mwindow = mwindow;
	set_checked(mwindow->every_frame);
}
EveryFrame::~EveryFrame()
{
}
int EveryFrame::handle_event()
{
	mwindow->every_frame ^= 1;
	set_checked(mwindow->every_frame);
}




Close::Close(MainWindow *mwindow)
 : BC_MenuItem("Close", "w", 'w')
{
	this->mwindow = mwindow;
}

Close::~Close()
{
}

int Close::handle_event()
{
	mwindow->gui->unlock_window();
	mwindow->stop_playback();
	mwindow->gui->lock_window();
	mwindow->close_file();
}



MainCanvas::MainCanvas(MainWindow *mwindow, int x, int y, int w, int h)
 : BC_Canvas(x, y, w, h, BLACK)
{
	this->mwindow = mwindow;
}

MainCanvas::~MainCanvas()
{
}

int MainCanvas::notify_load()
{
	set_color(WHITE);
	set_font(LARGEFONT);
	draw_center_text(get_w() / 2, get_h() / 2, "LOADING...", LARGEFONT);
	set_font(MEDIUMFONT);
	flash();
}

int MainCanvas::unnotify_load()
{
}


MainScrollbar::MainScrollbar(MainWindow *mwindow, int x, int y, int w)
 : BC_ISlider(x, y, w, BUTTONH, w, 0, 0, w, -1, -1, 1, 0)
{
	this->mwindow = mwindow;
}

MainScrollbar::~MainScrollbar()
{
}

int MainScrollbar::handle_event()
{
	if(mwindow->asset)
	{
		if(get_keypress())
		{
			if(get_keypress() == LEFT)
			{
				mwindow->gui->unlock_window();
				mwindow->stop_playback();
				mwindow->gui->lock_window();
				mwindow->frame_backward();
			}
			else
			if(get_keypress() == RIGHT)
			{
				mwindow->gui->unlock_window();
				mwindow->stop_playback();
				mwindow->gui->lock_window();
				mwindow->frame_forward();
			}
		}
		else
		{
			mwindow->gui->unlock_window();
			mwindow->stop_playback();
			mwindow->gui->lock_window();
			mwindow->current_sample = get_value();
			mwindow->flash_frame(toframes(get_value(), mwindow->samplerate, mwindow->frame_rate));
			mwindow->gui->canvas->flash();
		}
	}
}


PlayButton::PlayButton(MainWindow *mwindow, int x, int y)
 : BC_ForwardButton(x, y, BUTTONW, BUTTONH, 1)
{
	this->mwindow = mwindow;
}

PlayButton::~PlayButton()
{
}
int PlayButton::handle_event()
{
	if(get_mode() == 2)
	{
		mwindow->gui->unlock_window();
		mwindow->stop_playback();
		mwindow->gui->lock_window();
		update(1);
	}
	else
	{
		update(2);
		mwindow->start_playback();
	}
}

int PlayButton::keypress_event()
{
	if(get_keypress() == ' ') { handle_event(); return 1; }
	return 0;
}


FrameBackButton::FrameBackButton(MainWindow *mwindow, int x, int y)
 : BC_LeftTriangleButton(x, y, BUTTONW, BUTTONH, 1)
{
	this->mwindow = mwindow;
}

FrameBackButton::~FrameBackButton()
{
}
int FrameBackButton::handle_event()
{
	mwindow->gui->unlock_window();
	mwindow->stop_playback();
	mwindow->gui->lock_window();
	mwindow->frame_backward();
}

int FrameBackButton::keypress_event()
{
	if(get_keypress() == LEFT) { handle_event(); return 1; }
	return 0;
}

FrameForwardButton::FrameForwardButton(MainWindow *mwindow, int x, int y)
 : BC_RightTriangleButton(x, y, BUTTONW, BUTTONH, 1)
{
	this->mwindow = mwindow;
}

FrameForwardButton::~FrameForwardButton()
{
}
int FrameForwardButton::handle_event()
{
	mwindow->gui->unlock_window();
	mwindow->stop_playback();
	mwindow->gui->lock_window();
	mwindow->frame_forward();
}

int FrameForwardButton::keypress_event()
{
	if(get_keypress() == RIGHT) { handle_event(); return 1; }
	return 0;
}



LoadThread::LoadThread(MainWindow *mwindow)
 : Thread()
{
	this->mwindow = mwindow;
}
LoadThread::~LoadThread()
{
}
void LoadThread::run()
{
	int result = 0;
	{
		LoadDialog dialog(mwindow);
		dialog.create_objects();
		result = dialog.run_window();
		strcpy(mwindow->default_path, dialog.get_filename());
	}

	if(!result) 
	{
		mwindow->gui->lock_window();
		result = mwindow->load_file(mwindow->default_path);
		mwindow->gui->unlock_window();
	}
}


LoadDialog::LoadDialog(MainWindow *mwindow)
 : BC_FileBox("", 
 			mwindow->default_path, 
 			"Load", 
 			"Select the file to load from", 0, 0)
{
	this->mwindow = mwindow;
}
LoadDialog::~LoadDialog()
{
}
int LoadDialog::ok_event()
{
	set_done(0);
}
int LoadDialog::cancel_event()
{
	set_done(1);
}


ErrorThread::ErrorThread() : Thread()
{
}

ErrorThread::~ErrorThread()
{
}

int ErrorThread::run_error(char *text1, char *text2, char *text3, char *text4)
{
	this->text1 = text1;
	this->text2 = text2;
	this->text3 = text3;
	this->text4 = text4;
	Thread::synchronous = 0;
	startup_lock.lock();
	start();
	startup_lock.lock();
	startup_lock.unlock();
}

void ErrorThread::run()
{
	ErrorBox box("", "XMovie: FYI");
	box.create_objects(text1, text2, text3, text4);
	startup_lock.unlock();
	box.run_window();
}



