#include "settings.h"
#include "asset.h"
#include "mainwindow.h"

SettingsThread::SettingsThread(MainWindow *mwindow)
 : Thread()
{
	this->mwindow = mwindow;
	thread_running = 0;
}
SettingsThread::~SettingsThread()
{
}

void SettingsThread::run()
{
	if(thread_running) return;
	thread_running = 1;

	aspect_w = mwindow->aspect_w;
	aspect_h = mwindow->aspect_h;
	smp = mwindow->smp;
	square_pixels = mwindow->square_pixels;
	prebuffer_size = mwindow->prebuffer_size;

	SettingsWindow window(mwindow, this);
	window.initialize();
	this->window = &window;
	int result = window.run_window();
	if(!result)
	{
		if(mwindow->prebuffer_size != prebuffer_size || mwindow->smp != smp)
		{
			mwindow->error_thread->run_error("You must restart XMovie for the changes", "to take effect.");
		}

		if(aspect_w <= 0) aspect_w = 1;
		if(aspect_h <= 0) aspect_h = 1;
		mwindow->aspect_w = aspect_w;
		mwindow->aspect_h = aspect_h;
		mwindow->smp = smp;
		mwindow->square_pixels = square_pixels;
		mwindow->prebuffer_size = prebuffer_size;

		if(mwindow->asset && mwindow->asset->video_data)
		{
			int w = mwindow->asset->width, h = mwindow->asset->height;
			mwindow->fix_size(w, h, mwindow->gui->zoom_factor);
			mwindow->gui->resize_canvas(w, h);
		}
		else
			mwindow->gui->resize_canvas(mwindow->gui->canvas->get_w(), mwindow->gui->canvas->get_h());

		mwindow->flash_frame(toframes(mwindow->current_sample, mwindow->samplerate, mwindow->frame_rate));
		mwindow->gui->canvas->flash();
	}

// Lock out changes to the framerate
	change_lock.lock();
	this->window = 0;
	change_lock.unlock();
	thread_running = 0;
}

int SettingsThread::update_framerate()
{
	change_lock.lock();
	if(window)
	{
		char string[1024];
		sprintf(string, "%.02f", mwindow->actual_framerate);
		window->lock_window();
		window->frame_rate->update(string);
		window->unlock_window();
	}
	change_lock.unlock();
}



SettingsWindow::SettingsWindow(MainWindow *mwindow, SettingsThread *thread)
 : BC_Window("XMovie: Settings Ratio", 200, 265, 200, 265)
{
	this->mwindow = mwindow;
	this->thread = thread;
}
SettingsWindow::~SettingsWindow()
{
}
	
int SettingsWindow::initialize()
{
	char string[1024];
	int x, y;
	x = 10;
	y = 10;
	add_tool(new BC_Title(x, y, "Aspect Ratio"));
	y += 20;
	sprintf(string, "%.0f", thread->aspect_w);
	add_tool(new SettingsW(thread, x, y, string));
	x += 60;
	add_tool(new BC_Title(x, y, ":"));
	x += 10;
	sprintf(string, "%.0f", thread->aspect_h);
	add_tool(new SettingsH(thread, x, y, string));
	y += 40;
	x = 10;
	add_tool(new SettingsSquare(thread, x, y));
	y += 30;
	add_tool(new SettingsSMP(thread, x, y));
	y += 30;
	add_tool(new BC_Title(x, y, "Preload size"));
	y += 20;
	sprintf(string, "%ld", thread->prebuffer_size);
	add_tool(new SettingsPrebuff(thread, x, y, string));
	y += 40;
	add_tool(new BC_Title(x, y, "Actual framerate:"));
	sprintf(string, "%.2f", mwindow->actual_framerate);
	add_tool(frame_rate = new BC_Title(x + 130, y, string, MEDIUMFONT, RED));
	y += 40;

	add_tool(new SettingsOK(x, y));
	x += 100;
	add_tool(new SettingsCancel(x, y));
}

SettingsW::SettingsW(SettingsThread *thread, int x, int y, char *string)
 : BC_TextBox(x, y, 50, string)
{
	this->thread = thread;
}
SettingsW::~SettingsW()
{
}
	
int SettingsW::handle_event()
{
	thread->aspect_w = atof(get_text());
}

SettingsH::SettingsH(SettingsThread *thread, int x, int y, char *string)
 : BC_TextBox(x, y, 50, string)
{
	this->thread = thread;
}
SettingsH::~SettingsH()
{
}
	
int SettingsH::handle_event()
{
	thread->aspect_h = atof(get_text());
}


SettingsPrebuff::SettingsPrebuff(SettingsThread *thread, int x, int y, char *string)
 : BC_TextBox(x, y, 150, string)
{
	this->thread = thread;
}
SettingsPrebuff::~SettingsPrebuff()
{
}
	
int SettingsPrebuff::handle_event()
{
	if(atol(get_text()) >= 0)
		thread->prebuffer_size = atol(get_text());
}


SettingsSquare::SettingsSquare(SettingsThread *thread, int x, int y)
 : BC_CheckBox(x, y, 16, 16, thread->square_pixels, "Disable aspect ratio")
{
	this->thread = thread;
}
SettingsSquare::~SettingsSquare()
{
}
	
int SettingsSquare::handle_event()
{
	thread->square_pixels = get_value();
}


SettingsSMP::SettingsSMP(SettingsThread *thread, int x, int y)
 : BC_CheckBox(x, y, 16, 16, thread->smp, "Enable SMP extensions")
{
	this->thread = thread;
}
SettingsSMP::~SettingsSMP()
{
}

int SettingsSMP::handle_event()
{
	thread->smp = get_value();
}


SettingsOK::SettingsOK(int x, int y)
 : BC_BigButton(x, y, "OK")
{
}
SettingsOK::~SettingsOK()
{
}
int SettingsOK::handle_event()
{
	set_done(0);
}

int SettingsOK::keypress_event()
{
	if(get_keypress() == 13) { set_done(0); trap_keypress(); return 1; }
	return 0;
}

SettingsCancel::SettingsCancel(int x, int y)
 : BC_BigButton(x, y, "Cancel")
{
}
SettingsCancel::~SettingsCancel()
{
}
int SettingsCancel::handle_event()
{
	set_done(1);
}

int SettingsCancel::keypress_event()
{
	if(get_keypress() == ESC) { set_done(1); trap_keypress(); return 1; }
	return 0;
}
